from typing import List, Optional, Set, Tuple

from bech32 import bech32_decode, bech32_encode, convertbits

from .exceptions import InvalidLnurl, InvalidUrl


def _bech32_decode(bech32: str, *, allowed_hrp: Optional[Set[str]] = None) -> Tuple[str, List[int]]:
    hrp, data = bech32_decode(bech32)

    if not hrp or not data or (allowed_hrp and hrp not in allowed_hrp):
        raise ValueError(f"Invalid data or Human Readable Prefix (HRP): {hrp}.")

    return hrp, data


def _lnurl_clean(lnurl: str) -> str:
    lnurl = lnurl.strip()
    return lnurl.replace("lightning:", "") if lnurl.startswith("lightning:") else lnurl


def _lnurl_decode(lnurl: str) -> str:
    """
    Decode a LNURL and return a url string without performing any validation on it.
    Use `lnurl.decode()` for validation and to get `Url` object.
    """
    _, data = _bech32_decode(_lnurl_clean(lnurl), allowed_hrp={"lnurl"})

    try:
        bech32_data = convertbits(data, 5, 8, False)
        assert bech32_data
        url = bytes(bech32_data).decode("utf-8")
        return url
    except UnicodeDecodeError:  # pragma: nocover
        raise InvalidLnurl


def _url_encode(url: str) -> str:
    """
    Encode a URL without validating it first and return a bech32 LNURL string.
    Use `lnurl.encode()` for validation and to get a `Lnurl` object.
    """
    try:
        bech32_data = convertbits(url.encode("utf-8"), 8, 5, True)
        assert bech32_data
        lnurl = bech32_encode("lnurl", bech32_data)
    except UnicodeEncodeError:  # pragma: nocover
        raise InvalidUrl

    return lnurl.upper()



""" lnurl CLI """
import json
import sys

import click
import requests

from .core import encode as encode_lnurl
from .core import handle as handle_lnurl
from .types import Lnurl

# disable tracebacks on exceptions
sys.tracebacklimit = 0


@click.group()
def command_group():
    """
    Python CLI for LNURL
    decode and encode lnurls"""


@click.command()
@click.argument("url", type=str)
def encode(url):
    """
    encode a URL
    """
    encoded = encode_lnurl(url)
    click.echo(encoded.bech32)


@click.command()
@click.argument("lnurl", type=str)
def decode(lnurl):
    """
    decode a LNURL
    """
    decoded = Lnurl(lnurl)
    click.echo(decoded.url)


@click.command()
@click.argument("lnurl", type=str)
def handle(lnurl):
    """
    handle a LNURL
    """
    decoded = handle_lnurl(lnurl)
    click.echo(decoded.json())


@click.command()
@click.argument("lnurl", type=str)
@click.argument("amount", type=int)
def payment_request(lnurl, amount):
    """
    make a payment_request
    """
    res = handle_lnurl(lnurl)
    decoded = res.dict()

    if decoded["tag"] and decoded["tag"] == "payRequest":
        if decoded["minSendable"] <= amount <= decoded["maxSendable"]:
            res = requests.get(decoded["callback"] + "?amount=" + str(amount))
            res.raise_for_status()
            return click.echo(json.dumps(res.json()))
        else:
            click.echo("Amount not in range.")
    else:
        click.echo("Not a payRequest:")
    click.echo(res.json())


def main():
    """main function"""
    command_group.add_command(encode)
    command_group.add_command(decode)
    command_group.add_command(handle)
    command_group.add_command(payment_request)
    command_group()


if __name__ == "__main__":
    main()



# flake8: noqa

from .core import decode, encode, get, handle
from .models import (
    LnurlChannelResponse,
    LnurlErrorResponse,
    LnurlHostedChannelResponse,
    LnurlPayActionResponse,
    LnurlPayResponse,
    LnurlPayResponseComment,
    LnurlResponse,
    LnurlSuccessResponse,
    LnurlWithdrawResponse,
)
from .types import Lnurl



from typing import Any, Optional, Union

import requests
from pydantic import ValidationError

from .exceptions import InvalidLnurl, InvalidUrl, LnurlResponseException
from .helpers import _url_encode
from .models import LnurlAuthResponse, LnurlResponse, LnurlResponseModel
from .types import ClearnetUrl, DebugUrl, Lnurl, OnionUrl


def decode(bech32_lnurl: str) -> Union[OnionUrl, ClearnetUrl, DebugUrl]:
    try:
        return Lnurl(bech32_lnurl).url
    except (ValidationError, ValueError):
        raise InvalidLnurl


def encode(url: str) -> Lnurl:
    try:
        return Lnurl(_url_encode(url))
    except (ValidationError, ValueError):
        raise InvalidUrl


def get(url: str, *, response_class: Optional[Any] = None, verify: Union[str, bool] = True) -> LnurlResponseModel:
    try:
        r = requests.get(url, verify=verify)
        r.raise_for_status()
    except Exception as e:
        raise LnurlResponseException(str(e))

    if response_class:
        assert issubclass(response_class, LnurlResponseModel), "Use a valid `LnurlResponseModel` subclass."
        return response_class(**r.json())

    return LnurlResponse.from_dict(r.json())


def handle(
    bech32_lnurl: str, *, response_class: Optional[LnurlResponseModel] = None, verify: Union[str, bool] = True
) -> LnurlResponseModel:
    try:
        lnurl = Lnurl(bech32_lnurl)
    except (ValidationError, ValueError):
        raise InvalidLnurl

    if lnurl.is_login:
        return LnurlAuthResponse(callback=lnurl.url, k1=lnurl.url.query_params["k1"])

    return get(lnurl.url, response_class=response_class)
